/* -*-c-*- */
/* Copyright Olivier Chapuis (CNRS), Matthieu Langet (CNRS), 2005.
 *
 * <chapuis@lri.fr> <mathieu.langet@lri.fr>
 *
 * This software is a computer program whose purpose is to log the
 * window management activity into some files.
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 */


#include "config.h"

#include <string.h>

#include <X11/Xlib.h>

#include "include/wmstate.h"
#include "include/wmallowedactions.h"

#include "global.h"
#include "atoms.h"
#include "xget.h"

Window *virtualRoots = NULL;
unsigned int virtualRootsNbr = 0;

#define RootWin DefaultRootWindow(disp)

/*
 *
 * utils functions to get properties
 *
 */

void *get_property(
	Display *disp, Window win, atom_t atom, unsigned int *size)
{
	unsigned char *retval;
	Atom  type_ret;
	unsigned long  bytes_after, num_ret;
	long length;
	int  format_ret;
	void *data;
	int ok;

	retval = NULL;
	length = 0x7fffffff;
	ok = XGetWindowProperty(
		disp, win, atom.atom, 0, length, False, atom.type, &type_ret,
		&format_ret, &num_ret, &bytes_after, &retval);

	if ((ok == Success) && (retval) && (num_ret > 0) && (format_ret > 0))
	{
		data = malloc(num_ret * (format_ret >> 3));
		if (data)
		{
			memcpy(data, retval, num_ret * (format_ret >> 3));
		}
		XFree(retval);
		*size = num_ret * (format_ret >> 3);
		return data;
	}
	if (retval)
	{
		XFree(retval);
	}
	return NULL;
}

static
void *get_simple_property(
	Display *disp, Window win, atom_t atom)
{
	unsigned int size;

	return get_property(disp, win, atom, &size);
}

char *get_atoms_string(Display *disp, Atom *atoms, unsigned int size)
{
	char **str_list;
	char *str;
	int len = 0,i;

	if (atoms == NULL || size == 0)
	{
		return NULL;
	}
	
	str_list = (void *)malloc(sizeof(*str_list) * size);
	if (XGetAtomNames(disp, atoms, size, str_list) == 0)
	{
		free(str_list);
		return NULL;
	}
	for (i = 0; i < size; i++)
	{
		len += strlen(str_list[i]) + 1;
	}

	/* be large */
	str = malloc(len + 2*size + 2);
	str[0] = '\0';

	for (i = 0; i < size; i++)
	{
		strcat(str, str_list[i]);
		if (i != size-1)
		{
			strcat(str, " ");	
		}
		XFree(str_list[i]);
	}
	free(str_list);

	return str;
}

/*
 *
 * "check" stuff
 *
 */

static
Window get_wm_check_window(Display *disp, atom_t check_window_atom)
{
	Window rw;
	Window *wm_window, *check_window;
	Window junkRoot;
	int junkX, junkY;
	unsigned int junkWidth, junkHeight, junkBW, junkDepth;

	if ((wm_window = get_simple_property(
		     disp, RootWin, check_window_atom)) == NULL)
	{
		return None;
	}

	/* check that the window exist ... */	
	if (!XGetGeometry(
		    disp, wm_window[0], &junkRoot, &junkX, &junkY,
		    &junkWidth, &junkHeight, &junkBW,  &junkDepth))
	{
		return None;
	}

	/* check that the window is the good corresponding window */
	check_window = get_simple_property(
		disp, wm_window[0], check_window_atom);

	if (!check_window || check_window[0] != wm_window[0])
	{
		if (check_window)
		{
			free(check_window);
		}
		free(wm_window);
		return None;
	}

	rw = check_window[0];
	free(wm_window);
	free(check_window);

	return rw;
}

Window get_net_check_window(Display *disp)
{
	return get_wm_check_window(disp, atom_net_check);
}

Window get_win_check_window(Display *disp)
{
	return get_wm_check_window(disp, atom_win_check);
}

Atom *get_net_supported_atoms(Display *disp, unsigned int *rsize)
{
	Atom *atoms;
	unsigned int size;

	*rsize = 0;

	atoms = get_property(disp, RootWin, atom_net_supported, &size);
	*rsize = size/sizeof(Atom);
	return atoms;
}

Atom *get_win_supported_atoms(Display *disp, unsigned int *rsize)
{
	Atom *atoms;
	unsigned int size;

	*rsize = 0;

	atoms = get_property(disp, RootWin, atom_win_supported, &size);
	*rsize = size/sizeof(Atom);
	return atoms;
}

/*
 *
 * "root" properties
 *
 */

Window *get_client_list(Display *disp, unsigned int *ret_size)
{
	Window *client_list;
	unsigned int size;

	if ((client_list = get_property(
		     disp, RootWin, atom_client_list, &size)) == NULL)
	{
		return NULL;
	}

	*ret_size = size/sizeof(Window);
	return client_list;
}

Window *get_stacking_list(Display *disp, unsigned int *ret_size)
{
	Window *val;
	unsigned int size;

	*ret_size = 0;
	if ((val = get_property(
		     disp, RootWin, atom_net_client_list_stacking, &size)) ==
	    NULL)
	{
		return NULL;
	}
	*ret_size = size/sizeof(Window);
	return val;
}


Window get_active_window(Display *disp)
{
	Window *active_window;
	Window w = None;

	if ((active_window = get_simple_property(
		     disp, RootWin, atom_net_active_window)) != NULL)
	{
		w = *active_window;
		free(active_window);
	}

	return w;
}

unsigned long get_current_desktop(Display *disp)
{
	unsigned long *val;
	unsigned long desktop;

	if ((val = get_simple_property(
		     disp, RootWin, atom_net_current_desktop)) == NULL)
	{
		return 0;
	}

	desktop = val[0];
	free(val);

	return desktop;
}

long get_number_of_desktops(Display *disp)
{
	unsigned long *val;
	long nd;

	if ((val = get_simple_property(
		     disp, RootWin, atom_net_number_of_desktops)) == NULL)
	{
		return 0;
	}
	nd = val[0];
	free(val);

	return nd;
}

/* FIXME: a priori different desktop can have different viewport ??? */
Bool get_desktop_viewport(
	Display *disp, unsigned long *x, unsigned long *y)
{
	unsigned long *val;
	unsigned int size;

	if ((val = get_property(
		     disp, RootWin, atom_net_desktop_viewport, &size)) == NULL ||
	    size/sizeof(unsigned long) < 2)
	{
		if (val)
		{
			free(val);
		}
		return False;
	}
	*x = val[0];
	*y = val[1];
	free(val);

	return True;
}

/* FIXME: a priori different desktop can have different geometry ??? */
Bool get_desktop_geometry(
	Display *disp, unsigned long *w, unsigned long *h)
{
	unsigned long *val;
	unsigned int size;

	if ((val = get_property(
		     disp, RootWin, atom_net_desktop_geometry, &size)) == NULL ||
	    size/sizeof(unsigned long) < 2)
	{
		if (val)
		{
			free(val);
		}
		return False;
	}

	*w = val[0];
	*h = val[1];
	free(val);

	return True;
}

unsigned char **get_desktop_names(Display *disp, unsigned int *nbr)
{
	unsigned char *val,*tmp;
	unsigned char **names;
	unsigned int size;
	int count = 0, i = 0, j = 0, nn = 0;

	if ((val = get_property(
		     disp, RootWin, atom_net_desktop_names, &size)) == NULL)
	{
		*nbr = 0;
		return NULL;
	}

	for(i = 0; i < size; i++)
	{
		if (val[i] == 0)
		{
			nn++;
		}
	}
	if (nn == 0)
	{
		*nbr = 0;
		free(val);
		return NULL;
	}

	names = malloc(nn * sizeof(*names));

	tmp = val;
	for(i = 0; i < size; i++)
	{
		count++;
		if (val[i] == 0)
		{
			names[j] = malloc(count*sizeof(char));
			memcpy(names[j], tmp, count);
			j++;
			tmp = tmp + count;
			count = 0;
		}
	}
	*nbr = nn;

	free(val);
	return names;
}

Bool get_desktop_layout(Display *disp, desktop_layout_t *dl)
{
	unsigned long *val;
	unsigned int size;

	if ((val = get_property(
		     disp, RootWin, atom_net_desktop_layout, &size)) == NULL ||
	    size/sizeof(unsigned long) < 3)
	{
		if (val)
		{
			free(val);
		}
		return False;
	}

	dl->orientation = val[0];
	dl->columns = val[1];
	dl->rows = val[2];
	dl->starting_corner = 0;
	if (size/sizeof(unsigned long) > 3)
	{
		dl->starting_corner = val[3];
	}	
	free(val);

	return True;
}

Bool get_showing_desktop(Display *disp, unsigned long *bool)
{
	unsigned long *val;

	if ((val = get_simple_property(
		     disp, RootWin, atom_net_showing_desktop)) == NULL) 
	{
		return False;
	}
	*bool = val[0];
	free(val);
	return True;
}

Bool get_workarea(
	Display *disp, int *x, int *y, unsigned int *w, unsigned int *h)
{
	unsigned long *val;
	unsigned int size;

	if ((val = get_property(
		     disp, RootWin, atom_net_workarea, &size)) == NULL ||
	    size/sizeof(unsigned long) < 4)
	{
		if (val)
		{
			free(val);
		}
		return False;
	}

	*x = val[0];
	*y = val[1];
	*w = val[2];
	*h = val[3];
	free(val);

	return True;
}

/*
 *
 *  virtual root stuff
 *
 */

Bool isVirtualRoot(Display *disp, Window wind)
{
	int i;

	if (wind == RootWin)
	{
		return True;
	}

	for (i = 0; i < virtualRootsNbr; i++)
	{
		if (wind == virtualRoots[i])
		{
			return True;
		}
	}

	return False;
}

Bool get_set_virtual_roots(Display *disp)
{
	Window *val;
	unsigned int size;

	if ((val = get_property(
		     disp, RootWin, atom_net_virtual_roots, &size)) == NULL)
	{
		if (virtualRoots != NULL)
		{
			free(virtualRoots);
			virtualRoots = NULL;
			virtualRootsNbr = 0;
		}
		return False;
	}

	virtualRootsNbr = size/sizeof(Window);
	virtualRoots = val;

	return True;
}

/*
 *
 * "window" properties
 *
 */

unsigned long get_window_desktop(Display *disp, Window wind)
{
	unsigned long *val;
	unsigned long desktop;

	if ((val = get_simple_property(disp, wind, atom_wm_desktop)) == NULL)
	{
		return 0;
	}

	desktop = val[0];
	free(val);

	return desktop;
}

static
void get_state_from_net_wm_state(
	Display * disp, Atom *val, unsigned int size,
	unsigned long *state, int *layer)
{
	int i;
	*state = 0;
	*layer = -1;

	for(i = 0; i < size; i++)
	{
		if (val[i] == atom_net_wm_state_below.atom)
		{
			*layer = 2;
		}
		else if (val[i] == atom_net_wm_state_above.atom ||
			 val[i] == atom_net_wm_state_stays_on_top.atom)
		{
			*layer = 6;
		}
		else if (val[i] == atom_net_wm_state_fullscreen.atom)
		{
			*state |= WM_STATE_FULLSCREEN;
		}
		else if (val[i] == atom_net_wm_state_hidden.atom)
		{
			*state |= WM_STATE_MINIMIZED;
		}
		else if (val[i] == atom_net_wm_state_maximized_horz.atom ||
			 val[i] == atom_net_wm_state_maximized_horiz.atom)
		{
			*state |= WM_STATE_MAXIMIZED_HORIZ;
		}
		else if (val[i] == atom_net_wm_state_maximized_vert.atom)
		{
			*state |= WM_STATE_MAXIMIZED_VERT;
		}
		else if (val[i] == atom_net_wm_state_modal.atom)
		{
			*state = WM_STATE_MODAL;
		}
		else if (val[i] == atom_net_wm_state_shaded.atom)
		{
			*state |= WM_STATE_SHADED;
		}
		else if (val[i] == atom_net_wm_state_skip_pager.atom)
		{
			*state |= WM_STATE_SKIP_PAGER;
		}
		else if (val[i] == atom_net_wm_state_skip_taskbar.atom)
		{
			*state |= WM_STATE_SKIP_TASKBAR;
		}
		else if (val[i] == atom_net_wm_state_sticky.atom)
		{
			*state |= WM_STATE_STICKY;
		}
	}

	return;
}

void get_state_and_layer_window(
	Display * disp, Window wind, unsigned long *new_state, int *new_layer)
{
	unsigned long *pstate;
	Atom *val;
	unsigned int size;
	unsigned long state = 0;
	int layer = -1;

	if ((val = get_property(disp, wind, atom_net_wm_state, &size)) != NULL)
	{
		size = size/sizeof(Atom);
		get_state_from_net_wm_state(
			disp, val, size, &state, &layer);
		free(val);
	}

	if ((pstate = get_simple_property(disp, wind, atom_win_state)) != NULL)
	{
		/* our state correspond to the WIN state */
		state |= *pstate;
		free(pstate);
	}

	/* if we have a layer in the _NET_WM_STATE check the WIN_LAYER */
	if (layer != -1 &&
	    (pstate = get_simple_property(disp, wind, atom_win_layer)) != NULL)
	{
		layer = *pstate;
		free(pstate);
	}

#define WIN_HINTS_SKIP_FOCUS      (1<<0) /*"alt-tab" skips this win*/
#define WIN_HINTS_SKIP_WINLIST    (1<<1) /*do not show in window list*/
#define WIN_HINTS_SKIP_TASKBAR    (1<<2) /*do not show on taskbar*/
#define WIN_HINTS_GROUP_TRANSIENT (1<<3) /*Reserved - definition is unclear*/
#define WIN_HINTS_FOCUS_ON_CLICK  (1<<4) /*app only accepts focus if clicked*/
	if ((pstate = get_property(disp, wind, atom_win_hints, &size)) != NULL)
	{
		unsigned long tmp_state = *pstate;

		if (tmp_state & WIN_HINTS_SKIP_FOCUS)
		{
			state |= WM_STATE_SKIP_FOCUS;
		}
		if (tmp_state & WIN_HINTS_SKIP_FOCUS)
		{
			state |= WM_STATE_SKIP_WINLIST;
		}
		if (tmp_state & WIN_HINTS_SKIP_TASKBAR)
		{
			state |= WM_STATE_SKIP_TASKBAR;
		}
		if (tmp_state & WIN_HINTS_GROUP_TRANSIENT)
		{
			state |= WM_STATE_GROUP_TRANSIENT;
		}
		if (tmp_state & WIN_HINTS_GROUP_TRANSIENT)
		{
			state |= WM_STATE_FOCUS_ON_CLICK;
		}
		free(pstate);
	}

	*new_state = state;
	*new_layer = layer;

	return;
}

void get_layer_window(
	Display * disp, Window wind, int *new_layer)
{
	int *player;

	*new_layer = -1;
	if ((player = get_simple_property(disp, wind, atom_win_layer)) != NULL)
	{
		*new_layer = *player;
		free(player);
	}
}

Bool get_wm_state(Display *disp, Window wind, unsigned long *state)
{
	unsigned long *val;
	if ((val = get_simple_property(disp, wind, atom_wm_state)) == NULL)
	{
		return False;
	}

	*state = val[0];
	free(val);

	return True;
}

Atom *get_net_window_type(Display *disp, Window wind, unsigned int *size)
{
	Atom *atoms;
	if ((atoms = get_property(
		     disp, wind, atom_net_wm_window_type, size)) == NULL)
	{
		return NULL;
	}

	*size = *size/sizeof(Atom);
	return atoms;
}

Atom *get_win_window_type(Display *disp, Window wind, unsigned int *size)
{
	Atom *atoms;
	if ((atoms = get_property(
		     disp, wind, atom_win_window_type, size)) == NULL)
	{
		return NULL;
	}

	*size = *size/sizeof(Atom);
	return atoms;
}

char *get_window_type_str(Display *disp, Window wind)
{
	Atom *net_atoms;
	Atom *win_atoms;
	unsigned int net_size,win_size,len = 0;
	char *net_str = NULL;
	char *win_str = NULL;
	char *str;

	if ((net_atoms = get_net_window_type(disp, wind, &net_size)) != NULL)
	{
		net_str = get_atoms_string(disp, net_atoms, net_size);
		if (net_str)
		{
			len += strlen(net_str);
		}
		free(net_atoms);
	}
	if ((win_atoms = get_win_window_type(disp, wind, &win_size)) != NULL)
	{
		win_str = get_atoms_string(disp, win_atoms, win_size);
		if (win_str)
		{
			len += strlen(win_str);
		}
		free(win_atoms);
	}

	str = malloc(sizeof(char) * (len + strlen(NO_TYPE) + 3));
	str[0] = '\0';

	if (net_str)
	{
		strcat(str, net_str);
		free(net_str);
	}
	if (net_str && win_str)
	{
		strcat(str, " ");
	}
	if (win_str)
	{
		strcat(str, win_str);
		free(win_str);
	}
	if (!net_str && !win_str)
	{
		strcat(str, NO_TYPE);
	}
	
	return str;
}

char *get_client_machine(Display *disp, Window win)
{
	char *name;

	name = get_simple_property(disp, win, atom_client_machine);
	return name;
}

unsigned long get_window_pid(Display *disp, Window win)
{
	unsigned long *val;
	unsigned long pid;

	val = get_simple_property(disp, win, atom_net_wm_pid);

	if (val == NULL)
	{
		val = get_simple_property(disp, win, atom_mmpl_pid);
	}

	if (val == NULL)
	{
		return 0;
	}

	pid = val[0];
	free(val);
	return pid;
}

strut_t *get_strut(Display *disp, Window wind)
{
	unsigned long *val;
	unsigned int size;
	strut_t *strut;

	val = get_property(disp, wind, atom_net_wm_strut, &size);

	if (val == NULL || size/sizeof(unsigned long) < 4)
	{
		if (val)
		{
			free(val);
		}
		return NULL;
	}

	strut = (strut_t *)malloc(sizeof(strut_t));
	strut->left = val[0];
	strut->right = val[1];
	strut->top = val[2];
	strut->bottom = val[3];

	free(val);

	return strut;
}

partial_strut_t *get_partial_strut(Display *disp, Window wind)
{
	unsigned long *val;
	unsigned int size;
	partial_strut_t *p_strut;

	val = get_property(disp, wind, atom_net_wm_partial_strut, &size);

	if (val == NULL || size/sizeof(unsigned long) < 12)
	{
		if (val)
		{
			free(val);
		}
		return NULL;
	}

	p_strut = (partial_strut_t *)malloc(sizeof(partial_strut_t));
	p_strut->left = val[0];
	p_strut->right = val[1];
	p_strut->top = val[2];
	p_strut->bottom = val[3];
	p_strut->start.left = val[4];
	p_strut->end.left = val[5];
	p_strut->start.right = val[6];
	p_strut->end.right = val[7];
	p_strut->start.top = val[8];
	p_strut->end.top = val[9];
	p_strut->start.bottom = val[10];
	p_strut->end.bottom = val[11];
	free(val);

	return p_strut;
}

strut_t *get_frame_extents(Display *disp, Window wind)
{
	unsigned long *val;
	unsigned int size;
	strut_t *strut;

	val = get_property(disp, wind, atom_net_wm_extents, &size);

	if (val == NULL || size/sizeof(unsigned long) < 4)
	{
		if (val)
		{
			free(val);
		}
		val = NULL;
	}

	if (val == NULL)
	{
		val = get_property(disp, wind, atom_kde_net_frame_strut, &size);
	}
	
	if (val == NULL || size/sizeof(unsigned long) < 4)
	{
		if (val)
		{
			free(val);
		}
		return NULL;
	}

	strut = (strut_t *)malloc(sizeof(strut_t));
	strut->left = val[0];
	strut->right = val[1];
	strut->top = val[2];
	strut->bottom = val[3];

	free(val);

	return strut;
}

Bool get_icon_geometry(
	Display *disp, Window wind,
	int *x, int *y, unsigned int *w, unsigned int *h)
{
	unsigned long *val;
	unsigned int size;

	if ((val = get_property(
		     disp, wind, atom_net_wm_icon_geometry, &size)) == NULL ||
	    size/sizeof(unsigned long) < 4)
	{
		if (val)
		{
			free(val);
		}
		*x = *y = *w = *h = 0;
		return False;
	}

	*x = val[0];
	*y = val[1];
	*w = val[2];
	*h = val[3];
	free(val);

	return True;
}

Bool get_handle_icons(Display *disp, Window wind)
{
	unsigned long *val;
	unsigned int size;

	if ((val = get_property(
		     disp, wind, atom_net_wm_handle_icons, &size)) == NULL) 
	{
		return False;
	}
	free(val);
	return True;
}

Bool get_allowed_actions(Display *disp, Window wind, unsigned long *allowed)
{
	Atom *atoms;
	unsigned int size;
	int i;

	*allowed = 0;

	atoms = get_property(disp, wind, atom_net_wm_allowed_actions, &size);
	size = size/sizeof(Atom);

	if (atoms == NULL)
	{
		*allowed = -1;
		return False;
	} 
	
	for(i = 0; i < size; i++)
	{
		if (atoms[i] == atom_net_wm_action_move.atom)
			*allowed |= WM_ACTION_MOVE_BITS;
		else if (atoms[i] == atom_net_wm_action_resize.atom)
			*allowed |= WM_ACTION_RESIZE_BITS;
		else if (atoms[i] == atom_net_wm_action_minimize.atom)
			*allowed |= WM_ACTION_MINIMIZE_BITS;
		else if (atoms[i] == atom_net_wm_action_shade.atom)
			*allowed |= WM_ACTION_SHADE_BITS; 
		else if (atoms[i] == atom_net_wm_action_stick.atom)
			*allowed |= WM_ACTION_STICK_BITS; 
		else if (atoms[i] == atom_net_wm_action_maximize_horz.atom)
			*allowed |= WM_ACTION_MAXIMIZE_HORZ_BITS;
		else if (atoms[i] == atom_net_wm_action_maximize_vert.atom)
			*allowed |= WM_ACTION_MAXIMIZE_VERT_BITS; 
		else if (atoms[i] == atom_net_wm_action_fullscreen.atom)
			*allowed |= WM_ACTION_FULLSCREEN_BITS;
		else if (atoms[i] == atom_net_wm_action_change_desktop.atom)
			*allowed |= WM_ACTION_CHANGE_DESKTOP_BITS;
		else if (atoms[i] == atom_net_wm_action_close.atom)
			*allowed |= WM_ACTION_CLOSE_BITS; 
	}

	free(atoms);

	return True;
}

Window get_transientFor(Display *disp, Window w)
{
	Window tw = None;

	if (XGetTransientForHint(disp, w, &tw))
	{
		if (tw == None || tw == w)
		{
			tw = RootWin;
		}
	}
	return tw;
}

char *get_window_title(Display *disp, Window win, int *is_net)
{
	char *name;
	*is_net = 0;
	XTextProperty text_prop;
	char **list;
	int num;

	if (XGetWMName(disp, win, &text_prop) != 0 &&
	    text_prop.encoding == XA_STRING)
	{
		/* STRING encoding, use this as it is */
		name = (char *)text_prop.value;
		return name;
	}

	/* not STRING encoding, try to convert XA_COMPOUND_TEXT */
	if (XmbTextPropertyToTextList(disp, &text_prop, &list, &num) >= Success
	    && num > 0 && *list)
	{
		/* Does not consider the conversion is REALLY succeeded:
		 * XmbTextPropertyToTextList return 0 (== Success) on success,
		 * a negative int if it fails (and in this case we are not
		 * here), the number of unconvertible char on "partial"
		 * success*/
		XFree(text_prop.value); /* return of XGetWM(Icon)Name() */
		name = strdup(*list);
		/*ret_name->name_list = list;*/
	}
	else
	{
		name = (char *)text_prop.value;
	}
	if (list)
	{
		XFreeStringList(list);
	}
	if (name == NULL)
	{
		name = strdup("NoTitle");
	}
	else
	{
		/* ok */
	}
	return name;

#if 0
	char *title;
	unsigned int size;
	if ((name = get_property(
		     disp, win, atom_net_wm_visible_name, &size)) != NULL)
	{
		*is_net = 2;
	}
	else if ((name = get_property(
			  disp, win, atom_net_wm_name, &size)) != NULL)
	{
		*is_net = 1;
	}
	else if ((name = get_property(disp, win, atom_wm_name, &size)) != NULL)
	{
		/* ok, but in fact we should use XGetWMName() and handle
		 * text propetry (see libs/Flocale.c (FlocaleGetNameProperty) in
		 * the fvwm source code */
	}
	else
	{
		/* allocate for simplification (we can safely free) */
		name = strdup("NoTitle");
		size = 7;
	}

	title = malloc(size + 1);
	memcpy(title, name, size);
	title[size] = '\0';
	free(name);

	return title;
#endif
}

/*
 *
 * utils functions
 *
 */

Bool get_is_in_tree(Display *disp, Window start, Window to_found)
{
	Window junkroot;
	Window parent;
	Window *filles = NULL;
	int nfilles;
	Window frame = start;

	if (start == to_found)
	{
		return True;
	}
	while(XQueryTree(disp, frame, &junkroot, &parent, &filles, &nfilles) &&
	      !isVirtualRoot(disp, parent))
	{
		frame = parent;
		XFree(filles);
		filles = NULL;
		if (frame == to_found)
		{
			return True;
		}
	}

	if (filles)
	{
		XFree(filles);
	}

	if (frame == to_found)
	{
		return True;
	}

	return False;
}

Window get_toplevel_window(Display *disp, Window wind)
{
	Window junkroot;
	Window parent;
	Window *filles = NULL;
	int nfilles;
	Window toplevel = wind;

	while(XQueryTree(disp, toplevel, &junkroot, &parent, &filles, &nfilles)
	      && !isVirtualRoot(disp, parent))
	{
		toplevel = parent;
		XFree(filles);
		filles = NULL;
	}

	if (filles)
	{
		XFree(filles);
	}

	return toplevel;
}

Bool get_client_window(Display *disp, Window wind, Window *tl)
{
	Window junkroot;
	Window parent;
	Window *filles = NULL;
	int nfilles,i;
	unsigned long s;

	if (get_wm_state(disp, wind, &s))
	{
		*tl = wind;
		return True;
	}

	if (!XQueryTree(disp, wind, &junkroot, &parent, &filles, &nfilles))
	{
		if (filles)
			XFree(filles);
		return False;
	}

	for(i = 0; i < nfilles; i++)
	{
		if (get_wm_state(disp, filles[i], &s))
		{
			*tl = filles[i];
			if (filles)
				XFree(filles);
			return True;
		}
	}
	for(i = 0; i < nfilles; i++)
	{
		if (get_client_window(disp, filles[i], tl))
		{
			if (filles)
				XFree(filles);
			return True;
		}
	}
	if (filles)
		XFree(filles);

	return False;
}

Bool get_frame_geometry(
	Display *disp, Window wind, int *fx, int *fy, unsigned int *fw,
	unsigned int *fh, Window *frame, strut_t **frame_extents)
{
	int x, y;
	unsigned int w, h, bw, depth;
	Window junkroot;

	*frame = get_toplevel_window(disp, wind);

	if (!XGetGeometry(
		    disp, *frame, &junkroot, &x, &y, &w, &h, &bw, &depth))
	{
		/* no more window */
		return False;
	}

	*fx = x;
	*fy = y;
	*fh = h;
	*fw = w;

	if (*frame == wind)
	{
		/**frame_extents = (strut_t *)calloc(0, sizeof(strut_t));*/
		*frame_extents =
			(strut_t *)malloc(sizeof(strut_t));
		(*frame_extents)->right = 0;
		(*frame_extents)->left = 0;
		(*frame_extents)->top = 0;
		(*frame_extents)->bottom = 0;
	}
	else
	{
		*frame_extents = get_frame_extents(disp, wind);
	}
	if (*frame_extents == NULL)
	{
		if (!XGetGeometry(
			    disp, wind, &junkroot, &x, &y, &w, &h, &bw, &depth))
		{
			/* no more window */
		}
		else
		{
			Window child;
			long int right, left, top, bottom;
			XTranslateCoordinates(
				disp, wind, RootWin, x, y, &x, &y, &child);
			right = (*fx + *fw) - (x + w);
			left = x - *fx;
			top = y - *fy;
			bottom = (*fy + *fh) - (y + h);

			/* handle shaded window */
			if (right < 0 && left >= 0)
			{
				right = *fw - left;
			}
			if (right >= 0 && left < 0)
			{
				left = *fw - right;
			}
			if (top < 0 && bottom >= 0)
			{
				top = *fh - bottom;
			}
			if (top >= 0 && bottom < 0)
			{
				bottom = *fh - top;
			}
			if (right < 0 || left < 0 || top < 0 || bottom < 0)
			{
				/* ?? */
			}
			else
			{
				*frame_extents =
					(strut_t *)malloc(sizeof(strut_t));
				(*frame_extents)->right = right;
				(*frame_extents)->left = left;
				(*frame_extents)->top = top;
				(*frame_extents)->bottom = bottom;
			}
		}
	}
	return True;
}

XineramaScreenInfo *get_xinerama_info(Display *disp, int *count)
{
	int dummy_rc;

	XineramaScreenInfo *info = NULL;
	dummy_rc = 0;

	if (!(FScreenHaveXinerama &&
	      XineramaQueryExtension(disp, &dummy_rc, &dummy_rc) &&
	      XineramaIsActive(disp)))
	{
		return NULL;
	}

	info = XineramaQueryScreens(disp, count);

	return info;
}
